import logging
from collections import namedtuple

from .exceptions import ScreenError
from .version import __title__

try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk

logger = logging.getLogger(__title__)

Coords = namedtuple('Coords', ('top', 'left', 'right', 'bottom'))


class Grabber(tk.Tk):
    WINDOW_COLOR = '#ffffff'
    WINDOW_ALPHA = 0.2

    RECTANGLE_COLOR = '#000000'

    @classmethod
    def select_area(cls):
        logger.info("Selecting an area.")

        # run TK app to select an area
        scope = {'coords': None}

        def set_coords(_coords):
            scope['coords'] = _coords
            grabber.exit()  # close the window

        grabber = cls(on_selected=set_coords)
        try:
            grabber.mainloop()
        except KeyboardInterrupt:
            grabber.exit()

        coords = scope['coords']
        if not coords:
            raise ScreenError("Aborted!")

        # normalize coords
        coords = tuple(map(int, coords))
        left, right = sorted(coords[0::2])
        top, bottom = sorted(coords[1::2])

        logger.debug("Selected area %s.", coords)
        return Coords(top, left, right, bottom)

    def __init__(self, on_selected):
        tk.Tk.__init__(self)

        self.title(__title__)

        self._on_selected = on_selected

        self._coords = None
        self._rect_id = None
        self._is_drawing = False

        self._canvas = None

        self.initialize_geometry()
        self.initialize_controls()

    def initialize_geometry(self):
        self.wait_visibility()
        self.attributes('-topmost', True)
        self.attributes('-fullscreen', True)
        self.attributes('-alpha', self.WINDOW_ALPHA)

    def initialize_controls(self):
        self._canvas = tk.Canvas(self, bg=self.WINDOW_COLOR, cursor='crosshair')
        self._canvas.pack(fill=tk.BOTH, expand=1)

        self._canvas.bind('<Button-1>', self.start_drawing)
        self._canvas.bind('<Button-3>', self.exit)
        self._canvas.bind('<ButtonRelease-1>', self.stop_drawing)
        self._canvas.bind('<Motion>', self.draw_rectangle)

    def start_drawing(self, event):
        logger.debug("Selecting screen area.")
        self._is_drawing = True

        x = self._canvas.canvasx(event.x)
        y = self._canvas.canvasy(event.y)

        self._coords = [x, y, x, y]
        self._rect_id = self._canvas.create_rectangle(*self._coords, fill=self.RECTANGLE_COLOR)

    def draw_rectangle(self, event):
        if not self._is_drawing:
            return

        self._coords[2] = self._canvas.canvasx(event.x)
        self._coords[3] = self._canvas.canvasy(event.y)
        self._canvas.coords(self._rect_id, *self._coords)

    def stop_drawing(self, event):
        logger.debug("Screen area has been selected. Coords: %s", self._coords)
        self._is_drawing = False

        self._canvas.delete(self._rect_id)

        self._on_selected(self._coords)

    def exit(self, event=None):
        self.attributes('-alpha', 0)
        self.destroy()



import io
import logging
import struct
import zlib
from collections import namedtuple

from mss import mss

__all__ = ['capture_image']

logger = logging.getLogger(__name__)

Image = namedtuple('Image', ('stream', 'format'))


def capture_image(coords):
    """Take a screenshot."""
    logger.info("Capturing a screen.")
    screen = mss()
    mss_image = screen.grab({
        'top': coords.top,
        'left': coords.left,
        'width': coords.right - coords.left + 1,
        'height': coords.bottom - coords.top + 1,
    })

    stream = to_png(mss_image.rgb, mss_image.size)

    return Image(stream, 'png')


def to_png(data, size):
    """
    Dump data to a PNG file.

    This code is a part of the MSS library.
    Source: https://github.com/BoboTiG/python-mss/blob/d740fa774cae4b8ccaddf980cbf417ebe33117e7/mss/tools.py#L10
    MSS License: MIT
    2017-07-18

    :param bytes data: RGBRGB...RGB data.
    :param tuple size: The (width, height) pair.
    """

    width, height = size
    line = width * 3
    png_filter = struct.pack('>B', 0)
    scanlines = b''.join(
        [png_filter + data[y * line:y * line + line]
         for y in range(height)])

    magic = struct.pack('>8B', 137, 80, 78, 71, 13, 10, 26, 10)

    # Header: size, marker, data, CRC32
    ihdr = [b'', b'IHDR', b'', b'']
    ihdr[2] = struct.pack('>2I5B', width, height, 8, 2, 0, 0, 0)
    ihdr[3] = struct.pack('>I', zlib.crc32(b''.join(ihdr[1:3])) & 0xffffffff)
    ihdr[0] = struct.pack('>I', len(ihdr[2]))

    # Data: size, marker, data, CRC32
    idat = [b'', b'IDAT', zlib.compress(scanlines), b'']
    idat[3] = struct.pack('>I', zlib.crc32(b''.join(idat[1:3])) & 0xffffffff)
    idat[0] = struct.pack('>I', len(idat[2]))

    # Footer: size, marker, None, CRC32
    iend = [b'', b'IEND', b'', b'']
    iend[3] = struct.pack('>I', zlib.crc32(iend[1]) & 0xffffffff)
    iend[0] = struct.pack('>I', len(iend[2]))

    # store data in memory
    fileh = io.BytesIO()
    fileh.write(magic)
    fileh.write(b''.join(ihdr))
    fileh.write(b''.join(idat))
    fileh.write(b''.join(iend))
    fileh.seek(0)

    return fileh



import logging
import webbrowser

import pyperclip

logger = logging.getLogger(__name__)


def open_path(path):
    """Open a local path or URL."""
    logger.info('Opening %s.', path)
    webbrowser.open(path)


def copy_to_clipboard(text):
    """Writes a text to clipboard."""
    logger.info('Copying text to clipboard: %s', text)
    pyperclip.copy(text)



import logging
import os
from datetime import datetime

import requests
from requests.auth import HTTPDigestAuth

from .base import BaseStorage, File
from ..conf import config
from ..exceptions import StorageError

logger = logging.getLogger(__name__)


class Storage(BaseStorage):
    cloudapp_base_url = 'https://my.cl.ly/v3'

    def __init__(self):
        username = config.CLOUDAPP_USERNAME
        password = config.CLOUDAPP_PASSWORD
        if not username or not password:
            raise StorageError("Set 'cloudapp_username' and 'cloudapp_password' configs.")

        self._cloudapp_auth = HTTPDigestAuth(username, password)
        self._session = requests.Session()
        self._session.headers = {
            'Accept': 'application/json',
        }

    def save_image(self, image):
        logger.info('Uploading an image to CloudApp.')
        now = datetime.now()
        filename = 'Image %s.%s' % (now.strftime('%Y-%m-%d at %H:%M:%S'), image.format)

        file_detail = self._upload_file(image.stream, filename)

        return File(File.IMAGE, file_detail['share_url'])

    def _upload_file(self, stream, filename):
        cl_file_meta_url = self._build_cloudapp_url('items')
        cl_file_meta = self._request('POST', cl_file_meta_url, auth=self._cloudapp_auth, data={'name': filename})

        s3_url = cl_file_meta['url']
        s3_data = cl_file_meta['s3']
        s3_files = {'file': (filename, stream)}
        file_detail = self._request('POST', s3_url, data=s3_data, files=s3_files)

        return file_detail

    def _request(self, method, url, **kwargs):
        try:
            logger.debug('Sending a %s request to %s: %s', method, url, kwargs)
            response = self._session.request(method, url, **kwargs)
            logger.debug('Response from %s: status=%s body=%s', url, response.status_code, response.text)
        except requests.RequestException as e:
            logger.error(e)
            raise StorageError("Can't send a request to url: {}".format(e))

        self._handle_bad_request(response)

        return response.json()

    def _build_cloudapp_url(self, path):
        if path.startswith('/'):
            path = path[1:]

        return os.path.join(self.cloudapp_base_url, path)

    def _handle_bad_request(self, response):
        if response.ok:
            return

        if response.status_code >= 500:
            error_msg = "CloudApp broken :("
        elif response.status_code == 400:
            error_msg = "CloudApp doesn't want to accept our request. Sorry :("
        elif response.status_code in (401, 403):
            error_msg = "Invalid credentials for CloudApp."
        else:
            error_msg = "Bad response from Cloudapp ({}): {}".format(response.status_code, response.text)

        raise StorageError(error_msg)



import logging.config
import os
from shutil import copyfile

try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

logger = logging.getLogger(__name__)

DEFAULT_LEVEL = 'WARNING'
LOGGER_LEVELS = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET')


class Config(object):
    _default_section = 'app'
    _default_config_file_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'default.ini'))

    def __init__(self, auto_save=True, path=None):
        if not path:
            if 'GRAB_SCREEN_CONFIG_FILE' in os.environ:
                path = os.environ['GRAB_SCREEN_CONFIG_FILE']
            else:
                path = os.path.join(os.path.expanduser('~'), '.config', 'grab-screen', 'config.ini')

        self.auto_save = auto_save
        self.path = path
        self._parser = None

        self.load()

    def __iter__(self):
        for section in self._parser.sections():
            for option in self._parser.options(section):
                key = '{}_{}'.format(section, option).upper()
                yield key

    def __getattribute__(self, key):
        if key.isupper():
            return self.get(key)

        return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        if key.isupper():
            return self.set(key, value)

        return object.__setattr__(self, key, value)

    def __delattr__(self, key):
        if key.isupper():
            return self.delete(key)

        return object.__delattr__(self, key)

    def get(self, key):
        """Gets an option."""
        section, option = self._split_key(key)

        if not self._parser.has_option(section, option):
            return None

        return self._parser.get(section, option)

    def set(self, key, value):
        """Sets an option."""
        if value is None:
            self.delete(key)
            return

        section, option = self._split_key(key)

        self._parser.set(section, option, str(value))

        if self.auto_save:
            self.save()

    def delete(self, key):
        """Deletes an option."""
        section, option = self._split_key(key)
        removed = self._parser.remove_option(section, option)

        if removed and self.auto_save:
            self.save()

    def load(self):
        """Reads settings from the file."""
        if not os.path.exists(self.path):
            self.reset()
            return

        self._parser = ConfigParser()
        self._parser.read(self.path)

        self._load_logger()

    def save(self):
        """Writes settings to the file."""
        self._create_config_directory()

        for section in self._parser.sections():
            if not self._parser.options(section):
                self._parser.remove_section(section)

        with open(self.path, 'w') as config_file:
            self._parser.write(config_file)

    def reset(self):
        """Restores the default settings."""
        self._create_config_directory()

        copyfile(self._default_config_file_path, self.path)

        self.load()

    def _create_config_directory(self):
        directory = os.path.dirname(self.path)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def _load_logger(self):
        """
        Setup the logger configs.

        User can set `LOGGER_LEVEL` setting to change the logger level.
        """
        level = self.LOGGER_LEVEL
        if level:
            level = level.upper()
        if level not in LOGGER_LEVELS:
            logger.warning("Invalid logger level '%s'", level)
            level = DEFAULT_LEVEL

        logging.config.dictConfig({
            'version': 1,
            'formatters': {
                'simple': {
                    'format': '%(asctime)s [%(levelname)s] %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'simple',
                },
            },
            'loggers': {
                'grab_screen': {
                    'level': level,
                    'handlers': ['console'],
                },
            }
        })

    def _split_key(self, key):
        """Gets a section and an option names by the key."""
        key = key.lower()

        try:
            section, option = key.split('_', 1)
        except (ValueError, IndexError):
            section = self._default_section
            option = key

        if not self._parser.has_section(section):
            self._parser.add_section(section)

        return section, option



import importlib
import logging

from ..conf import config
from ..exceptions import StorageError

logger = logging.getLogger(__name__)


def get_storage(name=None):
    if not name:
        name = config.APP_STORAGE

    module_path = 'grab_screen.storages.{}'.format(name)

    try:
        module_ = importlib.import_module(module_path)
        return module_.Storage()
    except (ImportError, AttributeError):
        logger.error("Cannot find storage '%s' by path '%s", name, module_path)
        raise StorageError("Storage '%s' doesn't exist." % name)



import os
import sys

from setuptools import Command, find_packages, setup


def parse_requirements(path):
    with open(path) as file:
        requirements = tuple(filter(lambda line: line and not line.startswith('#'), file))

    return requirements


class LintCommand(Command):
    """Support setup.py lint."""

    description = 'Check the code style.'
    user_options = []

    @staticmethod
    def status(msg):
        """Prints things in bold."""
        sys.stdout.write('\033[1m{0}\033[0m\n'.format(msg))
        sys.stdout.flush()

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.status('Checking flake8 rules...')
        flake8_status = os.system('flake8')

        self.status('Checking bandit rules...')
        bandit_status = os.system('bandit -r .')

        status = 1 if flake8_status or bandit_status else 0
        sys.exit(status)


# load the description from the README file
with open('README.rst') as readme_file:
    long_description = readme_file.read()

# load the python dependencies
main_requirements = parse_requirements('requirements/main.txt')
test_requirements = parse_requirements('requirements/test.txt')

# load the package info
about = {}
with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'grab_screen', 'version.py'))) as version_file:
    exec(version_file.read(), about)

setup(
    name='grab-screen',
    description='Take screenshots and upload anywhere',
    long_description=long_description,
    version=about['__version__'],
    license='MIT',
    url='https://github.com/andrei-shabanski/grab-screen',
    author='Andrei Shabanski',
    author_email='andrei.shabanski@gmail.com',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,

    install_requires=main_requirements,
    tests_require=test_requirements,
    setup_requires=['pytest-runner'],
    extras_require={
        'dev': test_requirements,
    },

    entry_points={
        'console_scripts': [
            'grab-screen=grab_screen:main'
        ]
    },

    cmdclass={
        'lint': LintCommand,
    },

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Multimedia :: Graphics :: Capture',
        'Topic :: Multimedia :: Graphics :: Capture :: Screen Capture',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='screenshot image capture grab cloudapp'
)



class File(object):
    IMAGE = 'image'
    GIF = 'gif'
    VIDEO = 'video'

    def __init__(self, file_type, path):
        self.file_type = file_type
        self.path = path


class BaseStorage(object):
    """Base class for storages."""

    def save_image(self, image):
        """
        Saves an image.

        :param image: `Image` tuple with image data.
        :return: `File` object.
        """
        raise NotImplementedError()
